from .constants import S


def make_user_auth_service(context, request):
    return UserAuthService(request)


# https://pyramid-authsanity.readthedocs.io/en/latest/api/interfaces.html
class UserAuthService(object):

    def __init__(self, request):
        self.request = request
        self.session = session = request.session
        self.redis = session.redis
        self.user_definition = None

    def userid(self):
        if self.user_definition is None:
            raise Exception
        return self.user_definition.get('id')

    def groups(self):
        return self.user_definition.get('roles', [])

    def verify_ticket(self, principal, ticket):
        if not principal:
            return
        redis = self.redis
        redis_key = S['redis.users.prefix'] + principal
        if not redis.sismember(redis_key, ticket):
            return
        self.user_definition = self.session.get('user', {})

    def add_ticket(self, principal, ticket):
        redis = self.redis
        redis_key = S['redis.users.prefix'] + principal
        redis.sadd(redis_key, ticket)
        self.user_definition = self.request.user_definition

    def remove_ticket(self, ticket):
        try:
            principal = self.user_definition['id']
        except TypeError:
            return
        redis = self.redis
        redis_key = S['redis.users.prefix'] + principal
        redis.srem(redis_key, ticket)



def includeme(config):
    config.add_route(
        'authorizations.json',
        '/authorizations.json')
    config.add_route(
        'authorizations_enter',
        '/authorizations/enter/{provider_name}')
    config.add_route(
        'authorizations_enter_callback',
        '/authorizations/enter/{provider_name}/callback')
    config.add_route(
        'authorizations_leave',
        '/authorizations/leave')



from base64 import b64decode
from miscreant.aes.siv import SIV

from .constants import S


def encrypt(text):
    return S['crypt'].seal(text.encode('utf-8'))


def decrypt(text):
    return S['crypt'].open(text).decode('utf-8')


def get_crypt():
    key = b64decode(S['secret'])
    return SIV(key)



from invisibleroads_macros_configuration import Settings
from invisibleroads_macros_log import get_log


DEFAULT_SECRET = 'ESwOdBHt5L6lRNLHLwsFVy6gjqFmlx6r5E9seaqlra4='


S = Settings({
    'mock': True,
    'storage': 'session',
    'cookie_secure': False,
    'cookie_httponly': True,
    'verify_tls': True,
    'auth_state_length': 64,
    'public_attributes': ['name', 'email', 'image_url'],
    'image_url': (
        'http://3.bp.blogspot.com/_D_Z-D2tzi14/TMkHEi7p6CI/AAAAAAAAEA4/'
        'UyhnLtpt4hM/s400/drunk11.png'),
    'secret': DEFAULT_SECRET,
    'default_permission': 'see',
    'require_csrf': True,
    'redis.users.prefix': 'user.',
})
L = get_log(__name__)


INVISIBLEROADS_USERS_SETTINGS_PREFIX = 'invisibleroads_users.'


REDIS_SESSIONS_SETTINGS_PREFIX = 'redis.sessions.'
REDIS_SESSIONS_SECRET_LENGTH = 128
REDIS_SESSIONS_TIMEOUT_IN_SECONDS = 43200  # 12 hours
REDIS_SESSIONS_COOKIE_NAME = 's'
REDIS_SESSIONS_PREFIX = 'session.'


USER_DEFINITION = {
    'email': 'user@example.com',
}
DEFAULT_SECRET_ERROR_MESSAGE = f'''
!!! DEFAULT_SECRET is being used !!!

python -c "
from miscreant.aes.siv import SIV
from base64 import b64encode
print(b64encode(SIV.generate_key()).decode('utf-8'))"

{INVISIBLEROADS_USERS_SETTINGS_PREFIX}secret = YOUR-SECRET
'''.strip()



from invisibleroads_records.models import (
    CreationMixin,
    ModificationMixin,
    RecordMixin)
from sqlalchemy import Column
from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy.types import LargeBinary, Unicode

from .routines import decrypt, encrypt


class CaseInsensitiveString(str):

    def __lt__(self, other):
        return self.casefold() < other.casefold()

    def __le__(self, other):
        return self.casefold() <= other.casefold()

    def __eq__(self, other):
        return self.casefold() == other.casefold()

    def __ne__(self, other):
        return self.casefold() != other.casefold()

    def __gt__(self, other):
        return self.casefold() > other.casefold()

    def __ge__(self, other):
        return self.casefold() >= other.casefold()


class CaseInsensitiveEncryptComparator(Comparator):

    def operate(self, op, other, **kwargs):
        if hasattr(other, '__iter__'):
            other = [_.casefold() for _ in other]
        else:
            other = other.casefold()
        return op(self.__clause_element__(), encrypt(other), **kwargs)


class EmailMixin(object):
    'Mixin class for a case-insensitive encrypted email address'
    _email = Column(LargeBinary)

    @hybrid_property
    def email(self):
        return CaseInsensitiveString(decrypt(self._email))

    @email.setter
    def email(self, email):
        self._email = encrypt(email.casefold())

    @email.comparator
    def email(Class):
        return CaseInsensitiveEncryptComparator(Class._email)


class UserMixin(EmailMixin, ModificationMixin, CreationMixin, RecordMixin):
    __tablename__ = 'user'
    email = Column(Unicode, unique=True)
    name = Column(Unicode)
    image_url = Column(Unicode)



from os.path import abspath, dirname, join
from setuptools import find_packages, setup


ENTRY_POINTS = '''
'''
APPLICATION_CLASSIFIERS = [
    'Programming Language :: Python',
    'Framework :: Pyramid',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
    'License :: OSI Approved :: MIT License',
]
APPLICATION_REQUIREMENTS = [
    # web
    'pyramid',
    # database
    'redis',
    'sqlalchemy',
    # architecture
    'invisibleroads-posts >= 0.7.14',
    'invisibleroads-records >= 0.5.8.1',
    # security
    'miscreant',
    'pyramid-authsanity',
    'pyramid-redis-sessions',
    'requests-oauthlib',
    # shortcut
    'invisibleroads-macros-configuration >= 1.0.8',
    'invisibleroads-macros-log >= 1.0.3',
    'invisibleroads-macros-security >= 1.0.1',
    # other
]
TEST_REQUIREMENTS = [
    'pytest-cov',
]
FOLDER = dirname(abspath(__file__))
DESCRIPTION = '\n\n'.join(open(join(FOLDER, x)).read().strip() for x in [
    'README.md', 'CHANGES.md'])


setup(
    name='invisibleroads-users',
    version='0.6.1',
    description='Web application security defaults',
    long_description=DESCRIPTION,
    long_description_content_type='text/markdown',
    classifiers=APPLICATION_CLASSIFIERS,
    author='Roy Hyunjin Han',
    author_email='rhh@crosscompute.com',
    url='https://github.com/invisibleroads/invisibleroads-users',
    keywords='web wsgi bfg pylons pyramid invisibleroads',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    extras_require={'test': TEST_REQUIREMENTS},
    install_requires=APPLICATION_REQUIREMENTS,
    entry_points=ENTRY_POINTS)



from pytest import fixture

from . import models as M


@fixture
def users_request(records_request, users_config):
    User = M.User
    users_request = records_request
    users_request.__class__.authenticated_user = property(
        lambda self: User.get(self.database, self.authenticated_userid))
    yield users_request
    users_request.session.redis.flushall()


@fixture
def users_config(config):
    config.include('invisibleroads_users')
    yield config


@fixture
def user(users_request):
    database = users_request.database
    User = M.User
    user = User.make_unique_record(database)
    user.name = 'User'
    user.email = 'user@example.com'
    return user
